#!/usr/bin/env python

from efivaraccess import EfiVariables
import csv
import datetime
import glob
import os
import struct
import time

# logging
import logging
logging.basicConfig(level=logging.INFO, format="%(levelname)s: %(message)s")

# GUID definitions, see  InjectReportPkg/InjectReportPkg.dec
efinject_magic_guid = "8b1cd116-7f14-4594-8474-edbebda318ce"
efinject_vendor_guid = "1280f587-3e1c-4034-8d08-9e2a7d434213"

# format definitions for unpacking protocol data, in struct's format
format_id = "Q"
format_time = "HBBBBBxLHBx"
format_rdtsc = "Q"

# loggers active
loggers = (format_id, format_time, format_rdtsc)

# time between tick calls (seconds)
cycle_duration = 0.500

# struct information
struct_format = "=" + "".join(loggers)
struct_size = struct.calcsize(struct_format)

# efi variable reader
vars = EfiVariables.get_instance()


def int_or_str(s):
    try:
        return int(s)
    except ValueError:
        return s


def find_old_log():
    filenames = glob.glob("efinject-*.log")
    if filenames:
        filenames.sort()
        return filenames[-1]
    return None


def decode_from_buffer(buffer, index):
    return struct.unpack(struct_format,
                         buffer[struct_size*index:struct_size*(index+1)])


def complete_old_log():
    oldlog_filename = find_old_log()
    if not oldlog_filename:
        logging.info("Previous log file was not found, log completion not possible.")
        return
    
    with file(oldlog_filename, "r") as oldlog_file:
        oldlog_csv = csv.reader(oldlog_file)
        rows = [row for row in oldlog_csv]
    
    if not rows:
        logging.info("No data in previous log file, log completion not possible.")
        return
    
    pivot = tuple(int_or_str(element) for element in rows[-1])
    logging.debug("Reference row from log file: %r", pivot)
    
    old_entries = vars.read("EFInjectLastLog", efinject_vendor_guid)
    if not old_entries:
        logging.info("No last log in NVRAM found, log completion not possible.")
        return

    for i in range(len(old_entries) / struct_size):
        data = decode_from_buffer(old_entries, i)
        
        logging.debug("Comparing with last log entry: %r", data)
        if data == pivot:
            # found overlapping data
            logging.debug("Overlapping found at last log entry %i", i)
            start_copy_at = i + 1
            break
    else:
        # no overlapping data found, return
        logging.info("No overlapping data found, aborting log completion.")
        return
    
    n = 0
    with file(oldlog_filename, "a") as oldlog_file:
        oldlog_csv = csv.writer(oldlog_file)
        for i in range(start_copy_at, len(old_entries) / struct_size):
            data = decode_from_buffer(old_entries, i)
            logging.debug("Recovering additional entry: %r", data)
            oldlog_csv.writerow(data)
            n += 1
    
    logging.info("Log completion OK, recovered %d entries from NVRAM.", n)


def injection_loop():
    filename = "efinject-" + datetime.datetime.now().isoformat("_")[:19].translate(None, "-:") + ".log"
    logfile = file(filename, "a")
    logcsv = csv.writer(logfile)
    
    try:
        while True:
            # wait for next cycle
            time.sleep(cycle_duration)
            # perform tick, then read new log entries
            vars.read("Tick", efinject_magic_guid)
            new_entries = vars.read("ReadData", efinject_magic_guid)
            assert len(new_entries) % struct_size == 0
            # convert binary log entries into text for csv
            for i in range(len(new_entries) / struct_size):
                data = struct.unpack(struct_format, new_entries[struct_size*i:struct_size*(i+1)])
                logcsv.writerow(data)
            # force data to disk
            logfile.flush()
            os.fsync(logfile.fileno())
    
    except KeyboardInterrupt:
        # it's ok to quit by pressing Ctrl+C
        pass


def main():
    complete_old_log()
    
    if not vars.read("Test", efinject_magic_guid):
        logging.warning("InjectReport driver was not found. Exiting.")
        return
    
    injection_loop()

if __name__ == '__main__':
    main()
